@@ -12,6 +12,8 @@
 #include "passwordfield.h"
 #include "plasma_nm_openconnect.h"
 
+#include <QAtomicPointer>
+#include <QByteArray>
 #include <QComboBox>
 #include <QCryptographicHash>
 #include <QDialog>
@@ -27,6 +29,9 @@
 #include <QPushButton>
 #include <QTimer>
 #include <QWaitCondition>
+#include <QWebEngineCookieStore>
+#include <QWebEngineProfile>
+#include <QWebEngineView>
 
 #include <KLocalizedString>
 
@@ -84,6 +89,7 @@ public:
     int passwordFormIndex;
     QByteArray tokenMode;
     Token token;
+    QAtomicPointer<QSemaphore> waitForWebEngineFinish;
 
     enum LogLevels { Error = 0, Info, Debug, Trace };
 };
@@ -122,6 +128,7 @@ OpenconnectAuthWidget::OpenconnectAuthWidget(const NetworkManager::VpnSetting::P
             QOverload<const QString &, const QString &, const QString &, bool *>::of(&OpenconnectAuthWorkerThread::validatePeerCert),
             this,
             &OpenconnectAuthWidget::validatePeerCert);
+    connect(d->worker, &OpenconnectAuthWorkerThread::openWebEngine, this, &OpenconnectAuthWidget::openWebEngine);
     connect(d->worker, &OpenconnectAuthWorkerThread::processAuthForm, this, &OpenconnectAuthWidget::processAuthForm);
     connect(d->worker, &OpenconnectAuthWorkerThread::updateLog, this, &OpenconnectAuthWidget::updateLog);
     connect(d->worker, QOverload<const QString &>::of(&OpenconnectAuthWorkerThread::writeNewConfig), this, &OpenconnectAuthWidget::writeNewConfig);
@@ -144,6 +151,11 @@ OpenconnectAuthWidget::OpenconnectAuthWidget(const NetworkManager::VpnSetting::P
 OpenconnectAuthWidget::~OpenconnectAuthWidget()
 {
     Q_D(OpenconnectAuthWidget);
+    QSemaphore *webEngineCancel =
+        d->waitForWebEngineFinish.fetchAndStoreRelaxed(nullptr);
+    if (webEngineCancel) {
+        webEngineCancel->release();
+    }
     d->userQuit = true;
     if (write(d->cancelPipes[1], "x", 1)) {
         // not a lot we can do
@@ -189,14 +201,20 @@ void OpenconnectAuthWidget::readConfig()
         openconnect_setup_csd(d->vpninfo, getuid(), 1, wrapper);
     }
     if (!dataMap[NM_OPENCONNECT_KEY_PROXY].isEmpty()) {
-        const QByteArray proxy = QFile::encodeName(dataMap[NM_OPENCONNECT_KEY_PROXY]);
+        const QByteArray proxy = dataMap[NM_OPENCONNECT_KEY_PROXY].toUtf8();
         openconnect_set_http_proxy(d->vpninfo, OC3DUP(proxy.data()));
     }
 #if OPENCONNECT_CHECK_VER(5, 8)
     if (!dataMap[NM_OPENCONNECT_KEY_USERAGENT].isEmpty()) {
-        const QByteArray useragent = QFile::encodeName(dataMap[NM_OPENCONNECT_KEY_USERAGENT]);
+        const QByteArray useragent = dataMap[NM_OPENCONNECT_KEY_USERAGENT].toUtf8();
         openconnect_set_useragent(d->vpninfo, OC3DUP(useragent.data()));
     }
+#endif
+#if OPENCONNECT_CHECK_VER(5, 5)
+    if (!dataMap[NM_OPENCONNECT_KEY_VERSION_STRING].isEmpty()) {
+        const QByteArray versionstring = dataMap[NM_OPENCONNECT_KEY_VERSION_STRING].toUtf8();
+        openconnect_set_version_string(d->vpninfo, OC3DUP(versionstring.data()));
+    }
 #endif
     if (!dataMap[NM_OPENCONNECT_KEY_USERCERT].isEmpty()) {
         const QByteArray crt = QFile::encodeName(dataMap[NM_OPENCONNECT_KEY_USERCERT]);
@@ -480,6 +498,87 @@ void OpenconnectAuthWidget::logLevelChanged(int newLevel)
     }
 }
 
+
+void OpenconnectAuthWidget::handleWebEngineCookie(const QNetworkCookie &cookie)
+{
+    Q_D(OpenconnectAuthWidget);
+    const char *cookiesArr[3] = {nullptr};
+
+    cookiesArr[0] = cookie.name().constData();
+    cookiesArr[1] = cookie.value().constData();
+
+#if OPENCONNECT_CHECK_VER(5, 7)
+    struct oc_webview_result res;
+    res.cookies = cookiesArr;
+    if (!openconnect_webview_load_changed(d->vpninfo, &res)) {
+        QSemaphore *waitForWebEngineFinish =
+            d->waitForWebEngineFinish.fetchAndStoreRelaxed(nullptr);
+        if (waitForWebEngineFinish) {
+            waitForWebEngineFinish->release();
+        }
+    }
+#endif
+}
+
+void OpenconnectAuthWidget::handleWebEngineUrl(const QUrl &url)
+{
+    Q_D(OpenconnectAuthWidget);
+    // Hack due to lack of NULL pointer check in AnyConnect sso_detect_done
+    // logic in libopenconnect.
+    const char *cookiesArr[1] = {nullptr};
+    QByteArray urlByteArray = url.toString().toLocal8Bit();
+
+#if OPENCONNECT_CHECK_VER(5, 7)
+    struct oc_webview_result res;
+    res.cookies = cookiesArr;
+    res.uri = urlByteArray.constData();
+    if (!openconnect_webview_load_changed(d->vpninfo, &res)) {
+        QSemaphore *waitForWebEngineFinish =
+            d->waitForWebEngineFinish.fetchAndStoreRelaxed(nullptr);
+        if (waitForWebEngineFinish) {
+            waitForWebEngineFinish->release();
+        }
+    }
+#endif
+}
+
+void OpenconnectAuthWidget::openWebEngine(const char *loginUri, QSemaphore *waitForWebEngineFinish)
+{
+    Q_D(OpenconnectAuthWidget);
+    d->waitForWebEngineFinish.storeRelease(waitForWebEngineFinish);
+    auto webEngineView = new QWebEngineView(this);
+    QWebEngineCookieStore *cookieStore = webEngineView->page()->profile()->cookieStore();
+
+    // Some VPN protocols depend on parsing HTTP response headers to complete
+    // authentication. However, QtWebEngine does not provide an interface for
+    // capturing HTTP response data. QtWebEngine currently offers the capability
+    // to intercept HTTP requests made by the browser instance using
+    // QWebEngineUrlRequestInterceptor, but there is no equivalent for HTTP
+    // response data.
+    //
+    // VPN protocols with SSO support that do not depend on HTTP response headers
+    //
+    //     - Cisco AnyConnect Protocol
+    //
+    // VPN protocols with SSO support that do depend on HTTP response headers
+    //
+    //     - Palo Alto Networks GlobalProtect Protocol
+    //
+    // FIXME Add HTTP response header handling when the QtWebEngine ecosystem
+    // adds support for HTTP response intercepting.
+    connect(webEngineView, &QWebEngineView::urlChanged, this, &OpenconnectAuthWidget::handleWebEngineUrl);
+    connect(cookieStore, &QWebEngineCookieStore::cookieAdded, this, &OpenconnectAuthWidget::handleWebEngineCookie);
+    cookieStore->loadAllCookies();
+
+    webEngineView->load(QUrl(loginUri, QUrl::TolerantMode));
+    // QWebEngineView sizeHint fails to size window correctly based on contents
+    // when QLayout::setSizeConstraint(QLayout::SetFixedSize) is set. Using same
+    // size as webkitgtk is set to in GNOME/NetworkManager-openconnect.
+    webEngineView->setFixedSize(640, 480);
+
+    d->ui.loginBoxLayout->addWidget(webEngineView);
+}
+
 void OpenconnectAuthWidget::addFormInfo(const QString &iconName, const QString &message)
 {
     Q_D(OpenconnectAuthWidget);
@@ -794,3 +893,5 @@ void OpenconnectAuthWidget::viewServerLogToggled(bool toggled)
         d->ui.serverLogBox->setSizePolicy(policy);
     }
 }
+
+#include "moc_openconnectauth.cpp"
